@@ -10,15 +10,15 @@ from inspect import signature, Parameter
 from pathlib import Path
 
 import pandas as pd
+from deprecation import deprecated
 from pyispace import train_is
 from pyispace.utils import scriptcsv
 
 from .classification import Classifier
 from .feature_selection import featfilt
-from .measures import Measures
+from .measures import ClassificationMeasures
 from .utils import get_param_names
 
-
 metadata_file = 'metadata.csv'
 options_file = 'options.json'
 ih_file = 'ih.csv'
@@ -32,15 +32,22 @@ _mcr_envvar_value = "/usr/local/MATLAB/MATLAB_Runtime/v98/runtime/glnxa64:" \
                     "/usr/local/MATLAB/MATLAB_Runtime/v98/extern/bin/glnxa64"
 
 
-def build_metadata(data, rootdir=' ', labels_col=None, save=True, return_ih=False, **kwargs):
-    measures = Measures(data, labels_col=labels_col)
-    classifiers = Classifier(data, labels_col=labels_col)
+def build_metadata(data, rootdir=' ', problem='classification', output_col=None, save=True, return_ih=False, **kwargs):
+    if problem == 'classification':
+        measures = ClassificationMeasures(data, labels_col=output_col)
+        learners = Classifier(data, labels_col=output_col)
+    elif problem == 'regression':
+        # measures = ...
+        # learners = ...
+        raise NotImplementedError("Regression problems not yet supported. Coming soon!")
+    else:
+        raise ValueError(f"Unknown type of problem: '{problem}'.")
 
     m_params = {k: kwargs.get(k) for k in get_param_names(measures.calculate_all)}
     df_measures = measures.calculate_all(**m_params)
 
-    c_params = {k: kwargs.get(k) for k in get_param_names(classifiers.run_all)}
-    df_algo = classifiers.run_all(**c_params)
+    c_params = {k: kwargs.get(k) for k in get_param_names(learners.run_all)}
+    df_algo = learners.run_all(**c_params)
 
     df_metadata = pd.concat([df_measures, df_algo], axis=1)
     n_inst = len(df_metadata)
@@ -54,7 +61,7 @@ def build_metadata(data, rootdir=' ', labels_col=None, save=True, return_ih=Fals
         else:
             warnings.warn("Invalid directory '{0}'. File will not be saved.".format(path))
     if return_ih:
-        ih_values = classifiers.estimate_ih()
+        ih_values = learners.estimate_ih()
         df_ih = pd.DataFrame({'instance_hardness': ih_values},
                              index=pd.Index(range(1, n_inst + 1), name=instances_index))
         if save:
@@ -120,7 +127,10 @@ def update_options(d1: dict, d2: dict):
     return d1
 
 
-def run_isa(rootdir: Path, settings=None, save_output=True):
+def run_isa(rootdir: Path, settings=None, save_output=True, rotation_adjust=True):
+    if rotation_adjust is None:
+        rotation_adjust = False
+
     if not rootdir.exists():
         raise NotADirectoryError(f"Invalid directory {repr(repr(rootdir))}.")
 
@@ -140,12 +150,15 @@ def run_isa(rootdir: Path, settings=None, save_output=True):
     opts = update_options(opts, settings)
 
     logging.getLogger('pyispace').setLevel(logging.DEBUG)
-    out = train_is(metadata, opts)
+    out = train_is(metadata, opts, rotation_adjust)
     if save_output:
         scriptcsv(out, rootdir)
     return out
 
 
+@deprecated(deprecated_in='1.7.4',
+            removed_in='2.0.0',
+            details="only used in module graphene, which was discontinued")
 def clear_cache():
     files = glob.glob(str(_cache_path / '*'))
     for f in files:
@@ -155,6 +168,9 @@ def clear_cache():
             pass
 
 
+@deprecated(deprecated_in='1.7.4',
+            removed_in='2.0.0',
+            details="only used in module graphene, which was discontinued")
 def _set_environment():
     # delete cached files from previous run
     clear_cache()
@@ -166,6 +182,9 @@ def _set_environment():
         os.environ[_mcr_envvar] = _mcr_envvar_value
 
 
+@deprecated(deprecated_in='1.7.4',
+            removed_in='2.0.0',
+            details="only used in module graphene, which was discontinued")
 def run_pipeline(df_data: pd.DataFrame, configurations: dict):
     logger = logging.getLogger(__name__)
     logging.getLogger().setLevel(logging.INFO)
@@ -200,6 +219,9 @@ def run_pipeline(df_data: pd.DataFrame, configurations: dict):
     logger.info("Instance Hardness analysis finished.")
 
 
+@deprecated(deprecated_in='1.7.4',
+            removed_in='2.0.0',
+            details="only used in module graphene, which was discontinued")
 def load_cached_files():
     df_data = pd.read_csv(_cache_path / 'data.csv')
     df_metadata = pd.read_csv(_cache_path / metadata_file, index_col=instances_index)
