@@ -1,34 +1,15 @@
 import itertools
 import logging
-import pandas as pd
+
 import numpy as np
+import pandas as pd
+from scipy.sparse.csgraph import minimum_spanning_tree
 from sklearn import tree
-from sklearn.neighbors import NearestNeighbors, DistanceMetric
 from sklearn.model_selection import GridSearchCV
 from sklearn.naive_bayes import GaussianNB
-from scipy.sparse.csgraph import minimum_spanning_tree
+from sklearn.neighbors import NearestNeighbors, DistanceMetric
 
-_measures_dict = {
-    'kDN': 'k_disagreeing_neighbors',
-    'DS': 'disjunct_size',
-    'DCP': 'disjunct_class_percentage',
-    'TD_P': 'tree_depth_pruned',
-    'TD_U': 'tree_depth_unpruned',
-    'CL': 'class_likeliood',
-    'CLD': 'class_likeliood_diff',
-    'MV': 'minority_value',
-    'CB': 'class_balance',
-    'N1': 'borderline_points',
-    'N2': 'intra_extra_ratio',
-    'LSC': 'local_set_cardinality',
-    'LSR': 'ls_radius',
-    'Harmfulness': 'harmfulness',
-    'Usefulness': 'usefulness',
-    'F1': 'f1',
-    'F2': 'f2',
-    'F3': 'f3',
-    'F4': 'f4'
-}
+from .base import BaseMeasures
 
 
 def minmax(f: np.ndarray, y: np.ndarray):
@@ -92,7 +73,7 @@ def gower_distance(X):
     return cumsum_dist
 
 
-class Measures:
+class ClassificationMeasures(BaseMeasures):
     """
     Hardness measures class. It provides separate methods to compute each measure
 
@@ -105,9 +86,31 @@ class Measures:
         automatically
     """
 
-    def __init__(self, data: pd.DataFrame, ccp_alpha=None, labels_col=None):
-        self.logger = logging.getLogger(__name__)
+    _measures_dict = {
+        'kDN': 'k_disagreeing_neighbors',
+        'DS': 'disjunct_size',
+        'DCP': 'disjunct_class_percentage',
+        'TD_P': 'tree_depth_pruned',
+        'TD_U': 'tree_depth_unpruned',
+        'CL': 'class_likeliood',
+        'CLD': 'class_likeliood_diff',
+        'MV': 'minority_value',
+        'CB': 'class_balance',
+        'N1': 'borderline_points',
+        'N2': 'intra_extra_ratio',
+        'LSC': 'local_set_cardinality',
+        'LSR': 'ls_radius',
+        'Harmfulness': 'harmfulness',
+        'Usefulness': 'usefulness',
+        'F1': 'f1',
+        'F2': 'f2',
+        'F3': 'f3',
+        'F4': 'f4'
+    }
 
+    logger = logging.getLogger(__name__)
+
+    def __init__(self, data: pd.DataFrame, ccp_alpha=None, labels_col=None):
         if labels_col is None:
             self.labels_col = data.columns[-1]
             self.y = data.iloc[:, -1]
@@ -142,21 +145,6 @@ class Measures:
 
         self.dot = None
 
-    def _call_method(self, name, **kwargs):
-        return getattr(self, name)(**kwargs)
-
-    def calculate_all(self, measures_list=None):
-        if measures_list is None:
-            measures_list = _measures_dict.keys()
-
-        results = {}
-        for k in measures_list:
-            self.logger.info(f"Calculating measure {repr(k)}")
-            results[k] = self._call_method(_measures_dict[k])
-
-        df_measures = pd.DataFrame(results)
-        return df_measures.add_prefix('feature_')
-
     def k_disagreeing_neighbors(self, k=5, distance='gower'):
         data = self.data.copy()
         if distance == 'gower':
