@@ -1,11 +1,15 @@
-import yaml
 import inspect
 import traceback
-from sklearn.pipeline import make_pipeline
-from sklearn.preprocessing import StandardScaler
+import warnings
+
+import pandas as pd
+import yaml
+from pandas.api.types import is_numeric_dtype, is_integer_dtype, is_float_dtype
 from sklearn.decomposition import PCA
 from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
 from sklearn.neighbors import NeighborhoodComponentsAnalysis
+from sklearn.pipeline import make_pipeline
+from sklearn.preprocessing import StandardScaler
 
 
 def load_yaml_file(path):
@@ -75,3 +79,13 @@ def print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, lengt
         print()
 
 
+def check_dtype_consistency(s: pd.Series, expected_type: str) -> bool:
+    if expected_type == 'classification':
+        if is_float_dtype(s):
+            return False
+    elif expected_type == 'regression':
+        if not is_numeric_dtype(s):
+            return False
+        elif is_integer_dtype(s):
+            warnings.warn("")
+    return True
