@@ -1,5 +1,6 @@
 import io
 import logging
+from functools import reduce
 from pathlib import Path
 
 import holoviews as hv
@@ -9,17 +10,18 @@ import panel as pn
 from bokeh.models import HoverTool
 from bokeh.themes.theme import Theme
 from holoviews import opts, dim, streams
-from holoviews.plotting.links import DataLink
+from pandas.api.types import is_integer_dtype, is_float_dtype
 from shapely.geometry import Point, Polygon, MultiPolygon
 from sklearn.decomposition import PCA
 
+from pyhard.context import Workspace
 from pyhard.utils import reduce_dim
 
+pd.options.plotting.backend = "plotly"
 
 dark_color = '#292929'
 light_color = '#ffffff'
-# footprint_colors = {'good': '#3F81BC', 'bad': 'gray', 'best': 'green'}
-footprint_colors = {'good': 'green', 'best': 'orange', 'bad': 'yellow'}
+footprint_colors = {'good': '#58D68D', 'best': '#9B59B6'}
 
 dark_theme = Theme(
     json={
@@ -139,21 +141,20 @@ _my_path = Path(__file__).parent
 
 
 class App:
-    def __init__(self, df_dataset: pd.DataFrame, df_metadata: pd.DataFrame,
-                 df_is: pd.DataFrame, df_footprint: pd.DataFrame, df_foot_perf=None):
+    _tabs_id = ['Instance Space', 'Footprint performance', 'Selection explorer']
+    _tabs_inner_id = ['Statistics', 'Features', 'Meta-features']
+
+    def __init__(self, workspace: Workspace):
         hv.extension('bokeh')
         hv.renderer('bokeh').theme = light_theme
-        self.tmpl = pn.Template(light_template)
 
         self.logger = logging.getLogger(__name__)
-        # Temporary solution for known warnings
-        logging.getLogger("param").setLevel(logging.ERROR)
+        self.workspace = workspace
 
-        self.df_original = df_dataset
-        self.df_metadata = df_metadata
-        self.df_is = df_is
-        self.df_footprint = df_footprint
-        self.df_foot_perf = df_foot_perf
+        df_dataset = workspace.data.copy()
+        df_metadata = workspace.extended_metadata
+        df_is = workspace.is_coordinates
+        df_footprint = workspace.footprints
 
         if len(df_dataset.columns) > 3:
             X = df_dataset.iloc[:, :-1]
@@ -161,22 +162,29 @@ class App:
 
             pca = PCA(n_components=2)
             X_embedded = pca.fit_transform(X)
+            # tsne = TSNE(n_components=2)
+            # X_embedded = tsne.fit_transform(X)
 
             df = pd.DataFrame(X_embedded, columns=[f'Component1',  # ({(100 * pca.explained_variance_ratio_[0]):.2f}%)
                                                    f'Component2'],  # ({(100 * pca.explained_variance_ratio_[1]):.2f}%)
                               index=X.index)
             df_dataset = pd.concat([df, y], axis=1)
-        self.df_dataset = df_dataset
 
-        data = self.df_is.join(self.df_dataset)
-        self.data = data.join(self.df_metadata)
+        data = df_is.join(df_dataset)
+        self.data = data.join(df_metadata)
         self.mlist = ['circle', 'triangle', 'square', 'diamond', 'asterisk', 'hex', '+', 'x']
-        self.is_kdims = df_is.columns.to_list()[0:2]
         self.data_dims = df_dataset.columns.to_list()
-        self.data_kdims = self.data_dims[0:2]
         self.class_label = self.data_dims[2]
         self.meta_dims = df_metadata.columns.to_list()
 
+        # Scatter kdims and vdims
+        is_cols = df_is.columns.to_list()
+        self.is_kdims = is_cols[0]
+        self.is_vdims = [is_cols[1], self.class_label] + self.meta_dims
+        self.data_kdims = self.data_dims[0]
+        self.data_vdims = [self.data_dims[1], self.class_label] + self.meta_dims
+
+        # Panel widgets
         self.w_color = pn.widgets.Select(options=self.meta_dims + [self.class_label], value=self.meta_dims[0])
         self.w_color_range = pn.widgets.RangeSlider(start=0, end=20, value=(0, 5), step=0.5)
         self.w_checkbox = pn.widgets.Checkbox(name='manual colorbar range', value=False)
@@ -187,13 +195,19 @@ class App:
 
         self.bbox = None
         self.cmap = 'coolwarm'
+        self._page_tabs = pn.Tabs(*[(name, None) for name in self._tabs_id], dynamic=False)
+        self._inner_tabs = pn.Tabs(*[(name, None) for name in self._tabs_inner_id], dynamic=True, tabs_location='left')
+        self._page_tabs[2] = (self._tabs_id[2], self._inner_tabs)
+        self.populate_tabs()
 
     @staticmethod
+    @DeprecationWarning
     def footprint2polygons(footprint: np.ndarray):
         poly_list = np.split(footprint, np.argwhere(np.isnan(footprint).any(axis=1)).flatten())
         return MultiPolygon(list(map(lambda x: Polygon(x[~np.isnan(x).any(axis=1)]), poly_list)))
 
     @classmethod
+    @DeprecationWarning
     def remove_intersection(cls, fp_good: np.ndarray, fp_bad: np.ndarray):
         p_good = MultiPolygon(cls.footprint2polygons(fp_good))
         p_bad = MultiPolygon(cls.footprint2polygons(fp_bad))
@@ -230,46 +244,34 @@ class App:
 
     def footprint_area(self, algo):
         try:
-            border_points_good = self.df_footprint.xs([algo, 'good']).values
+            border_points_good = self.workspace.footprints.xs([algo, 'good']).values
         except KeyError:
             border_points_good = np.array([[0, 0]])
-        # try:
-        #     border_points_bad = self.df_footprint.xs([algo, 'bad']).values
-        # except KeyError:
-        #     border_points_bad = np.array([[0, 0]])
         try:
-            border_points_best = self.df_footprint.xs([algo, 'best']).values
+            border_points_best = self.workspace.footprints.xs([algo, 'best']).values
         except KeyError:
             border_points_best = np.array([[0, 0]])
 
-        # if border_points_good.shape[0] > 2 and border_points_bad.shape[0] > 2:
-        #     try:
-        #         border_good, border_bad = self.remove_intersection(border_points_good, border_points_bad)
-        #     except TopologicalError:
-        #         border_good, border_bad = border_points_good, border_points_bad
-        #         self.logger.warning(f"It was not possible to remove overlapping regions from the {algo} footprints. "
-        #                             "Likely cause is invalidity of the geometry.")
-        # else:
-        #     border_good, border_bad = border_points_good, border_points_bad
         border_good, border_best = border_points_good, border_points_best
 
-        footprint_good = hv.Polygons([border_good.tolist()], label='Good').opts(line_width=1, line_alpha=0.2,
-                                                                                line_color='black',
-                                                                                fill_color=footprint_colors['good'],
-                                                                                fill_alpha=.2, show_legend=True)
-        # footprint_bad = hv.Polygons([border_bad.tolist()], label='Bad').opts(line_width=1, line_alpha=0.2,
-        #                                                                      line_color='black',
-        #                                                                      fill_color=footprint_colors['bad'],
-        #                                                                      fill_alpha=.2, show_legend=True)
-        footprint_best = hv.Polygons([border_best.tolist()], label='Best').opts(line_width=1, line_alpha=0.2,
-                                                                                line_color='black',
-                                                                                fill_color=footprint_colors['best'],
-                                                                                fill_alpha=.2, show_legend=True)
-        return footprint_good * footprint_best  # * footprint_bad
+        footprint_good = hv.Polygons([border_good.tolist()], label='Good Footprint').opts(line_width=1, line_alpha=0.2,
+                                                                                          line_color='black',
+                                                                                          fill_color=footprint_colors[
+                                                                                              'good'],
+                                                                                          fill_alpha=.2,
+                                                                                          show_legend=True)
+
+        footprint_best = hv.Polygons([border_best.tolist()], label='Best Footprint').opts(line_width=1, line_alpha=0.2,
+                                                                                          line_color='black',
+                                                                                          fill_color=footprint_colors[
+                                                                                              'best'],
+                                                                                          fill_alpha=.2,
+                                                                                          show_legend=True)
+        return footprint_good * footprint_best
 
     def select_instances(self):
         if self.bbox is None:
-            return pd.DataFrame()
+            return pd.Series(False, index=self.data.index)
         x, y = list(self.bbox.keys())
         if len(self.bbox[x]) == 2:
             V1 = np.column_stack([self.bbox[x], self.bbox[y]])
@@ -281,7 +283,7 @@ class App:
             contour = list(map(tuple, np.column_stack([self.bbox[x], self.bbox[y]])))
         polygon = Polygon(contour)
         mask = self.data[[x, y]].apply(lambda p: polygon.contains(Point(p[0], p[1])), raw=True, axis=1)
-        return self.df_original[mask]
+        return mask
 
     def data_space(self, c, lim, autorange_on):
         if not autorange_on:
@@ -290,12 +292,12 @@ class App:
         # hover_list = [c] + hover_list
         # tooltips = [(s, '@' + s) for s in hover_list]
         # hover = HoverTool(tooltips=tooltips)
-        scatter1 = hv.Scatter(self.data, kdims=self.data_kdims, vdims=[self.class_label] + self.meta_dims
-                              ).opts(responsive=True, aspect=1.2, color=c,
+        scatter1 = hv.Scatter(self.data, kdims=self.data_kdims, vdims=self.data_vdims
+                              ).opts(responsive=True, aspect=1.1, color=c,
                                      cmap=cmap, show_grid=True,
                                      marker=dim(self.class_label).categorize(self.mlist),
                                      tools=['lasso_select', 'box_select', 'hover'],
-                                     size=6, framewise=True, colorbar=True, clim=lim)
+                                     size=7, framewise=True, colorbar=True, clim=lim)
         return scatter1
 
     def instance_space(self, c, lim, autorange_on):
@@ -305,12 +307,12 @@ class App:
         # hover_list = [c] + hover_list
         # tooltips = [(s, '@' + s) for s in hover_list]
         # hover = HoverTool(tooltips=tooltips)
-        scatter2 = hv.Scatter(self.data, kdims=self.is_kdims, vdims=[self.class_label] + self.meta_dims
-                              ).opts(responsive=True, aspect=1.2, color=c,
+        scatter2 = hv.Scatter(self.data, kdims=self.is_kdims, vdims=self.is_vdims
+                              ).opts(responsive=True, aspect=1.1, color=c,
                                      cmap=cmap, show_grid=True,
                                      marker=dim(self.class_label).categorize(self.mlist),
                                      tools=['lasso_select', 'box_select', 'hover'],
-                                     size=6, framewise=True, colorbar=True, clim=lim)
+                                     size=7, framewise=True, colorbar=True, clim=lim)
         return scatter2
 
     def get_pane(self):
@@ -326,25 +328,22 @@ class App:
 
         def selection_callback1(bbox, region_element, selection_expr, resetting):
             self.bbox = bbox
-            # button.button_type = 'default'
-            # if bbox is not None:
-            #     button.disabled = False
             if resetting:
                 self.bbox = None
-                # button.disabled = True
+            self.populate_tabs()
             return hv.Polygons([[[0, 0]]])
 
         @pn.depends(footprint=self.w_footprint_algo.param.value, fp_on=self.w_footprint_on.param.value)
         def selection_callback2(bbox, region_element, selection_expr, footprint, fp_on):
             self.bbox = bbox
-            # button.button_type = 'default'
-            # if bbox is not None:
-            #     button.disabled = False
+
+            self.populate_tabs()
+
             if fp_on:
                 return self.footprint_area(footprint)
             else:
-                return hv.Polygons([[[0, 0]]], label='Good').opts(fill_color=footprint_colors['good']) * \
-                       hv.Polygons([[[0, 0]]], label='Best').opts(fill_color=footprint_colors['best'])
+                return (hv.Polygons([[[0, 0]]], label='Good Footprint').opts(fill_color=footprint_colors['good']) *
+                        hv.Polygons([[[0, 0]]], label='Best Footprint').opts(fill_color=footprint_colors['best']))
                 # hv.Polygons([[[0, 0]]], label='Bad').opts(fill_color=footprint_colors['bad']) * \
 
         dmap1 = hv.DynamicMap(update_plot1)
@@ -360,7 +359,8 @@ class App:
         sel2_dmap = hv.DynamicMap(selection_callback2, streams=[selection2])
 
         def file_cb():
-            df = self.select_instances()
+            mask = self.select_instances()
+            df = self.workspace.data[mask]
             sio = io.StringIO()
             df.to_csv(sio)
             sio.seek(0)
@@ -373,7 +373,7 @@ class App:
 
         layout = (dmap1 * sel1_dmap + dmap2 * sel2_dmap).cols(2).opts(
             opts.Layout(shared_axes=False, shared_datasource=True, framewise=True),
-            opts.Polygons(show_legend=True))
+            opts.Polygons(show_legend=True, legend_position='bottom'))
 
         gspec = pn.GridSpec(sizing_mode='stretch_both', background=light_color, margin=0)
         gspec[0, 0] = pn.Column('## Color', self.w_color, '### Color Bar', self.w_checkbox, self.w_color_range,
@@ -383,124 +383,148 @@ class App:
                                 '## Selection', button,
                                 background=light_color)
         gspec[0, 1:5] = layout
-        tabs = pn.Tabs(('PyHard', gspec))
-        tabs.append(('Footprint performance', pn.widgets.DataFrame(self.df_foot_perf, name='DataFrame',
-                                                                   disabled=True, sizing_mode='stretch_both')))
-        return tabs
+        self._page_tabs[0] = (self._tabs_id[0], gspec)
+        self._page_tabs[1] = (self._tabs_id[1], pn.widgets.DataFrame(self.workspace.footprint_performance,
+                                                                     name='Performance',
+                                                                     disabled=True,
+                                                                     sizing_mode='stretch_both'))
+        tmpl = pn.Template(light_template)
+        tmpl.add_panel(name='IS', panel=self._page_tabs)
+        return tmpl
 
-    def show(self, **kwargs):
-        tabs = self.get_pane()
-        self.tmpl.add_panel(name='IS', panel=tabs)
-        self.tmpl.servable()
-        self.tmpl.show(title="Instance Hardness", **kwargs)
+    def populate_tabs(self):
+        mask = self.select_instances()
+        df_selection = self.workspace.data[mask].copy()
+        output_col = df_selection.columns[-1]
+        df_selection.loc[:, output_col] = df_selection.iloc[:, -1].apply(lambda x: str(x))
+        cols = df_selection.columns[:-1].to_list()
 
-
-def app(df_dataset: pd.DataFrame, df_metadata: pd.DataFrame, df_is: pd.DataFrame, df_footprint: pd.DataFrame):
-    hv.extension('bokeh')
-    hv.renderer('bokeh').theme = dark_theme
-    tmpl = pn.Template(dark_template)
-
-    if len(df_dataset.columns) > 3:
-        X = df_dataset.iloc[:, :-1]
-        y = df_dataset.iloc[:, -1]
-        X_embedded = reduce_dim(X, y)
-        df = pd.DataFrame(X_embedded, columns=['V1', 'V2'], index=X.index)
-        df_dataset = pd.concat([df, y], axis=1)
-
-    data = df_is.join(df_dataset)
-    data = data.join(df_metadata)
-    mlist = ['circle', 'triangle', 'square', 'diamond', '+', 'x']
-    is_kdims = df_is.columns.to_list()[0:2]
-    data_dims = df_dataset.columns.to_list()
-    data_kdims = data_dims[0:2]
-    class_label = data_dims[2]
-    meta_dims = df_metadata.columns.to_list()
-
-    # data['class'] = data['class'].apply(lambda x: mlist[x])
-    def footprint_area(algo):
-        points_good = df_footprint.xs([algo, 'good']).values.tolist()
-        points_bad = df_footprint.xs([algo, 'bad']).values.tolist()
-        footprint_good = hv.Polygons([points_good], label='Good').opts(line_width=1, line_alpha=0.2, line_color='black',
-                                                                       fill_color='white', fill_alpha=.2,
-                                                                       show_legend=True)
-        footprint_bad = hv.Polygons([points_bad], label='Bad').opts(line_width=1, line_alpha=0.2, line_color='black',
-                                                                    fill_color='orange', fill_alpha=.2,
-                                                                    show_legend=True)
-        return footprint_good * footprint_bad
-
-    def plotter(c, lim, autorange_on, hover_list, footprint_selected, fp_on, **kwargs):
-        if not autorange_on:
-            lim = (np.nan, np.nan)
-        cmap = 'Blues_r'
-        scatter1 = hv.Scatter(data, kdims=data_kdims, vdims=[class_label, c] + meta_dims,
-                              label='Original Data').opts(responsive=True, aspect=1.2, color=c,
-                                                          cmap=cmap, show_grid=True,
-                                                          marker=dim(class_label).categorize(mlist))
-
-        scatter = hv.Scatter(data, kdims=is_kdims,
-                             vdims=[class_label, c] + meta_dims).opts(responsive=True,
-                                                                      aspect=1.2, color=c,
-                                                                      cmap=cmap, show_grid=True,
-                                                                      marker=dim(class_label).categorize(mlist))
-        if fp_on:
-            area = footprint_area(footprint_selected)
+        # selection dataframe tab
+        if not mask.any():
+            df_table = self.workspace.data[~mask].copy()
+            df_table[output_col] = df_table[output_col].apply(lambda x: str(x))
         else:
-            area = hv.Polygons([[[0, 0]]], label='Good').opts(show_legend=True) * \
-                   hv.Polygons([[[0, 0]]], label='Bad').opts(show_legend=True)
-            # scatter2 = hv.Scatter(data, kdims=is_kdims, vdims=[class_label, c] + meta_dims,
-            #                       label='Instance Space').opts(responsive=True, aspect=1.2, color=c,
-            #                                                    cmap=cmap, show_grid=True,
-            #                                                    marker=dim(class_label).categorize(mlist))
-        scatter2 = (scatter * area).relabel('Instance Space')
+            df_table = df_selection
+        desc_all = pn.widgets.DataFrame(df_table.describe(include='all'), name='Selection',
+                                        disabled=True, sizing_mode='stretch_both')
 
-        hover_list = [c] + hover_list
-        tooltips = [(s, '@' + s) for s in hover_list]
-        hover = HoverTool(tooltips=tooltips)
+        desc = df_table.groupby(output_col).describe(include='all').unstack(1)
+        desc.index = desc.index.set_names(['Feature', 'Statistic'], level=[0, 1])
+        desc = desc.reset_index(level=[0]).pivot(columns='Feature').reorder_levels([1, 0]).sort_index()
+        desc.columns = desc.columns.droplevel(0)
+        desc_class = pn.widgets.DataFrame(desc, name='Selection', disabled=True, sizing_mode='stretch_both',
+                                          show_index=True, hierarchical=True)
+        table = pn.Column(desc_all, "__Descriptive statistics by class__", desc_class, sizing_mode='stretch_width')
 
-        dlink = DataLink(scatter1, scatter2)
+        self._inner_tabs[0] = (self._tabs_inner_id[0], table)
 
-        return (scatter1 + scatter2).cols(2).opts(opts.Scatter(tools=['box_select', 'lasso_select', hover],
-                                                               size=4, framewise=True, colorbar=True, clim=lim),
-                                                  opts.Layout(shared_axes=True, shared_datasource=True,
-                                                              framewise=True))
+        # metadata selection tab
+        df_feat = self.workspace.metadata[mask].filter(regex='^feature_').melt(var_name='feature')
+        boxplot_feat = hv.BoxWhisker(df_feat, kdims='feature', vdims='value').opts(responsive=True,
+                                                                                   aspect=3,
+                                                                                   xrotation=45,
+                                                                                   show_grid=True,
+                                                                                   box_fill_color=dim('feature').str(),
+                                                                                   cmap='glasbey_cool',
+                                                                                   tools=['hover'])
 
-    w_color = pn.widgets.Select(options=meta_dims + [class_label], value='')
-    w_color_range = pn.widgets.IntRangeSlider(start=-40, end=40, value=(0, 30), step=1)
-    w_checkbox = pn.widgets.Checkbox(name='manual colorbar range', value=False)
-    w_selector_hover = pn.widgets.MultiChoice(value=data_dims + is_kdims,
-                                              options=data.columns.to_list())
-    w_footprint_on = pn.widgets.Checkbox(name='draw footprint area', value=False)
-    w_footprint_algo = pn.widgets.Select(options=df_footprint.index.unique(level='algo').to_list(), value='')
+        df_algo = self.workspace.metadata[mask].filter(regex='^algo_').melt(var_name='algorithm', value_name='logloss')
+        boxplot_algo = hv.BoxWhisker(df_algo, kdims='algorithm', vdims='logloss').opts(responsive=True,
+                                                                                       aspect=3,
+                                                                                       xrotation=45,
+                                                                                       show_grid=True,
+                                                                                       box_fill_color=dim(
+                                                                                           'algorithm').str(),
+                                                                                       cmap='glasbey_cool',
+                                                                                       tools=['hover'])
 
-    @pn.depends(color=w_color.param.value, lim=w_color_range.param.value, autorange_on=w_checkbox.param.value,
-                hover_list=w_selector_hover.param.value, footprint=w_footprint_algo.param.value,
-                fp_on=w_footprint_on.param.value)
-    def update_plot(color, lim, autorange_on, hover_list, footprint, fp_on, **kwargs):
-        return plotter(color, lim, autorange_on, hover_list, footprint, fp_on)
+        self._inner_tabs[2] = (self._tabs_inner_id[2],
+                               pn.Column('# Distribution of the meta-features', pn.Row(boxplot_feat),
+                                         '# Distribution of the algorithms', pn.Row(boxplot_algo),
+                                         sizing_mode='stretch_both'))
 
-    dmap = hv.DynamicMap(update_plot)
+        integer_cols = [col for col in cols if is_integer_dtype(df_selection[col])]
+        float_cols = [col for col in cols if is_float_dtype(df_selection[col])]
 
-    # row = pn.Row(pn.Column(pn.WidgetBox('## Color', w_color,
-    #                                     '### Color Bar', w_checkbox, w_color_range,
-    #                                     height=200, width=250, sizing_mode='scale_both'),
-    #                        sizing_mode='scale_both'), dmap, sizing_mode='scale_both')  # pn.layout.HSpacer()
-    # pane = pn.Column(row, '## Hover Info', w_selector_hover)
-    # pane.show(title="Instance Hardness")
+        w_var_box = pn.widgets.MultiSelect(name='Features', value=[], options=float_cols)
+        w_var_hist = pn.widgets.Select(name='Feature', options=cols)
+        w_var_bar = pn.widgets.Select(name='Feature', options=integer_cols)
 
-    md_color = '<span style="color:#292929">{0}</span>'
-    gspec = pn.GridSpec(sizing_mode='stretch_both', background='#292929', margin=0)
-    # gspec[0, 0:5] = pn.pane.Markdown('# Instance Hardness analysis app', style={'color': '#1A76FF'})
-    gspec[0:10, 0] = pn.Column('## Color', w_color,  # md_color.format('Color')
-                               '### Color Bar', w_checkbox, w_color_range,  # '### ' + , w_checkbox, w_color_range,
-                               pn.Row(pn.Spacer(), height=20),
-                               '## Footprint', w_footprint_on, w_footprint_algo,
-                               background='#292929')  # pn.Row(pn.Spacer(), height=20)
-    gspec[0:10, 1:5] = dmap
-    # gspec[10, 0:4] = pn.Row('## Hover options', w_selector_hover)
-    tmpl.add_panel(name='A', panel=gspec)
-    tmpl.servable()
-    tmpl.show(title="Instance Hardness")
-    # gspec.show(title="Instance Hardness")
+        @pn.depends(feat_list=w_var_box.param.value)
+        def make_box(feat_list):
+            return hv.BoxWhisker(df_selection[feat_list + [output_col]].melt(id_vars=output_col, var_name='feature'),
+                                 kdims=['feature', output_col],
+                                 vdims='value').opts(responsive=True,
+                                                     aspect=2.5,
+                                                     xrotation=45,
+                                                     show_grid=True,
+                                                     box_fill_color=dim(output_col).str(),
+                                                     box_cmap='Set1',
+                                                     tools=['hover'],
+                                                     fontscale=1.2)
+
+        @pn.depends(var=w_var_bar.param.value)
+        def make_bar(var):
+            if var is None:
+                bar = hv.Bars([])
+            else:
+                kdims = [var, output_col]
+                s = df_selection[kdims].value_counts()
+                s.name = 'Count'
+                s = s.to_frame()
+                bar = hv.Bars(s)
+            tooltips = [(output_col, f'@{output_col}'), ('Count', '@Count')]
+            bar.opts(responsive=True,
+                     aspect=2.5,
+                     multi_level=True,
+                     stacked=False,
+                     show_grid=True,
+                     fontscale=1.2,
+                     tools=[HoverTool(tooltips=tooltips)])
+            return bar
+
+        @pn.depends(var=w_var_hist.param.value)
+        def make_hist(var):
+            classes = df_selection.iloc[:, -1].unique()
+            hist = [hv.Histogram(np.histogram(df_selection[df_selection[output_col] == c][var]), label=str(c))
+                    for c in classes]
+            if len(hist) == 0:
+                hist = hv.Histogram([])
+            else:
+                hist = reduce(lambda h1, h2: h1 * h2, hist)
+            hist.opts('Histogram',
+                      responsive=True,
+                      aspect=2.5,
+                      alpha=0.8,
+                      show_grid=True,
+                      muted_fill_alpha=0.1,
+                      framewise=True,
+                      tools=['hover'],
+                      cmap='Set1',
+                      fontscale=1.2,
+                      xlabel=var)
+            return hist
+
+        dmap_box = hv.DynamicMap(make_box)
+        dmap_box.opts(framewise=True)
+
+        dmap_hist = hv.DynamicMap(make_hist)
+        dmap_hist.opts(framewise=True)
+
+        dmap_bar = hv.DynamicMap(make_bar)
+        dmap_bar.opts(framewise=True)
+
+        accordion = pn.Accordion(('Boxplot', pn.Row(w_var_box, dmap_box, sizing_mode='stretch_both')),
+                                 ('Histogram', pn.Row(w_var_hist, dmap_hist, sizing_mode='stretch_both')),
+                                 ('Barplot', pn.Row(w_var_bar, dmap_bar, sizing_mode='stretch_both')),
+                                 sizing_mode='stretch_both')
+
+        self._inner_tabs[1] = (self._tabs_inner_id[1], accordion)
+
+    def show(self, port=5001, **kwargs):
+        tmpl = self.get_pane()
+        pn.serve(tmpl, port=port, show=True, title='Instance Hardness',
+                 websocket_origin=[f'127.0.0.1:{port}', f'localhost:{port}'])
 
 
 class Demo:
@@ -600,7 +624,7 @@ class Demo:
         # dlink = DataLink(scatter1, scatter2)
 
         return (scatter1 + scatter2).opts(opts.Scatter(tools=['box_select', 'lasso_select', 'tap', hover],
-                                                       size=4, colorbar=True, clim=lim, framewise=True),
+                                                       size=6, colorbar=True, clim=lim, framewise=True),
                                           opts.Layout(shared_axes=False, shared_datasource=True)).cols(2)
 
     def update_components(self):
